/* @(#)sb.c	1.6	2/10/93 */
/*******************************************************************************

This code was written and designed by 
Andrew Gene HALL of I.N. Services Pty. Ltd., Scarborough WA, Australia,
for the 
University of Western Australia, Crawley WA, Australia.

*******************************************************************************/
/*
** Modified by David Argentar of BioMagResBank 28 July 1996, with
** permission of the authors
*/

#include <stdio.h>
#include <stdlib.h>
#include "utilitys.h"
#include "diagmesg.h"
#include "printsf.h"
#include "request_space.h"
#include "interpret.h"
#include "dictionary.h"
#include "sb_lib.h"

extern void request_restart();

/* macros and constants */
#define FNAME_BUFFER_BLOCK 4

/* globals */
int		ARGc;	/* global version of argc */
char		**ARGv;	/* global version of argv */
Block_Seq	*beginning_of_BS = NULL;
Block_Seq	*input_sf = NULL;
Request_Seq	*requests = NULL;
Boolean         strip_stream = FALSE; /* CHANGE 17/6 */
int             line_max = 512 ; /* CHANGE 17/6 */

/* parser stuff */
extern FILE		*star_in;
extern int 		parse_star();
extern FILE		*request_in;

/* library entry point to read and parse a star file */
void init_star(star_file_name)
char *star_file_name;
{

   /*error_fp = fopen("/dev/null", "w");*/
   current_level = ERROR;
   error_fp = stderr;
   input_sf = beginning_of_BS;

   if((star_in = fopen(star_file_name, "r")) == NULL) {
     /* error in opening the file */
     ERROR_MSG("Cannot open file for input:");
     ERROR_MSG(star_file_name);
   } 
   else {
     /* input file opened successfully */
     INFORM_MSG("Input from:");
     INFORM_MSG(star_file_name);
   }

   /* parse the input */
   ARGc = 0;
   if( parse_star() ) {
      ERROR_MSG("Could not parse the input.");
      exit(-1);
   }
   input_sf = beginning_of_BS;
   return;
}


/* library entry point to retrieve string values from a star file */

void star_get_string_tag(scope, tag_name, result, cnt)
char *scope, *tag_name;
char ***result;
int *cnt;
{
  FILE *tmp_request_fp = NULL;
  char tmp_request_fname[L_tmpnam];
  FILE *output_fp = stdout;
  Block_Seq *output_sf = NULL;
  static int first_pass = 1;
  char *arg_ptr;

  if( tmpnam(tmp_request_fname) ) {
    if((tmp_request_fp = fopen(tmp_request_fname, "w")) == NULL) {
      tmp_request_fp = NULL;
      Buffer_MSG( ERROR, "Cannot open temporary file for requests:");
      Buffer_MSG( ERROR, tmp_request_fname);
    }
  }
  else
    Buffer_MSG( ERROR, "Could not create a unique temporary filename.");

  if(!scope)
    ERROR_MSG("star_get_string_tag passed null pointer for scope");
  if(!tag_name)
    ERROR_MSG("star_get_string_tag passed null pointer for tag_name");
  if(!result)
    ERROR_MSG("star_get_string_tag passed null pointer for result");
  if(!cnt)
    ERROR_MSG("star_get_string_tag passed null pointer for cnt");
  if(!scope[0])
    ERROR_MSG("star_get_string_tag passed empty string for scope");
  if(!tag_name[0])
    ERROR_MSG("star_get_string_tag passed empty string for tag_name");

  if(!strncmp(scope, "save_",5))
    fprintf(tmp_request_fp, "if_ %s\nscope_save_frame_\n%s\nendscope_\nendif_\n",
      scope, tag_name);
  else
    if(!strncmp(scope, "data_",5))
      fprintf(tmp_request_fp, "if_ %s\nscope_data_block_\n%s\nendscope_\nendif_\n",
        scope, tag_name);
    else
      if(!strcmp(scope, "global_"))
        fprintf(tmp_request_fp, "if_ global_\n%s\nendif_\n", tag_name);
      else
      {
        ERROR_MSG("Invalid scope. ");
      }

  fclose(tmp_request_fp);

  /* parse and perform requests */
  /* attempt to open the request file */

  if((request_in = fopen(tmp_request_fname, "r")) == NULL) {
    ERROR_MSG("Cannot open file for input:");
    ERROR_MSG(tmp_request_fname);
  }
  else {
    /* input file opened successfully */
    INFORM_MSG("Input from:");
    INFORM_MSG(tmp_request_fname);
    ARGc = 0;
  }
  if(first_pass)
    first_pass = 0;
  else
  {
    arg_ptr = &(tmp_request_fname[0]);
    ARGv = &arg_ptr;
    ARGc = 1;
  }

  if( parse_request() ) {
    ERROR_MSG("Could not parse the requests.");
    exit(-1);
  }
  else
    /* perform the requests */
    output_sf = requests_froma_RS( requests, input_sf, NULL, input_sf);

  unlink(tmp_request_fname);
     
  /* write the output */
  fprint_Block_Seq(output_fp, output_sf, result, cnt);

  return;
}

/* library entry point to retrieve numeric values from a star file */
void star_get_numeric_tag(scope, tag_name, result, result_flags, cnt)
char *scope, *tag_name;
double **result;
int **result_flags;
int *cnt;
{
  char **sgst_result;
  char *last_char;
  int i, sgst_cnt, len;

  if(!scope)
    ERROR_MSG("star_get_numeric_tag passed null pointer for scope");
  if(!tag_name)
    ERROR_MSG("star_get_numeric_tag passed null pointer for tag_name");
  if(!result)
    ERROR_MSG("star_get_numeric_tag passed null pointer for result");
  if(!result_flags)
    ERROR_MSG("star_get_numeric_tag passed null pointer for result_flags");
  if(!cnt)
    ERROR_MSG("star_get_numeric_tag passed null pointer for cnt");
  if(!scope[0])
    ERROR_MSG("star_get_numeric_tag passed empty string for scope");
  if(!tag_name[0])
    ERROR_MSG("star_get_numeric_tag passed empty string for tag_name");

  star_get_string_tag(scope, tag_name, &sgst_result, &sgst_cnt);

  if(!sgst_cnt)
  {
    result = 0;
    result_flags = 0;
    cnt = 0;
    return;
  }

  *result = (double *)malloc(sizeof(double) * sgst_cnt);
  *result_flags = (int *)malloc(sizeof(int) * sgst_cnt);
  for(i = 0; i < sgst_cnt; i++)
  {
    if(!strcmp(sgst_result[i], "."))
    {
      (*result)[i] = 0.0;
      (*result_flags)[i] = NULL_FLAG;
      continue;
    }
    len = (int)strlen(sgst_result[i]);
    last_char = &(sgst_result[i][len - 1]);
    (*result)[i] = strtod(sgst_result[i], &last_char);
    if(last_char != &(sgst_result[i][len]))
    {
      (*result)[i] = 0.0;
      (*result_flags)[i] = ISNT_NUM_FLAG;
    }
    else
      (*result_flags)[i] = OK_FLAG;
  }

  free(sgst_result);
  (*cnt) = sgst_cnt;
  return;
}
